//====================================================================================//
//                                                                                    //
//                            Class Probability Vector                                //
//                                                                                    //
//====================================================================================//
//  This File:   vct.C                      Language: C++  (xlC and CC)               //
//  Software:    PEPS 3.1                    Advisor: Brigitte.Plateau@imag.fr        //
//  Error Codes: 2350                                                                 //
//====================================================================================//
//  Creation:    19/dec/95                        by: Paulo.Fernandes@imag.fr         //
//  Last Change: 13/feb/03                        by: Anne.Benoit@imag.fr             //
//====================================================================================//
#include "peps.h"

     //--------------------------------------------------------------------------//
     //                 static variable: The Probability Vector                  //
     //--------------------------------------------------------------------------//
VCT VCT::vct = VCT();

     //--------------------------------------------------------------------------//
     //               It initializes the states of 'this' - CND                  //
     //--------------------------------------------------------------------------//
     //    This function is used to initialize the elements of 'this' according  //
     // to one of the following options ('o'):                                   //
     //                                                                          //
     //   - 'eq_vec',      all reachable states are initialized with an uniform  //
     //                    distribution (equiprobable);                          //
     //   - 'ap_vec',      all reachable states are initialized with a           //
     //                    distribution proportional to the inverse of the       //
     //                    diagonal elements of the descriptor ('CND::cnd');     //
     //                                                                          //
     //   The size of the vector and the size of each automata is taken from the //
     // descriptor 'CND::cnd'. The field 'permutation' is set to the normal      //
     // order { 0, 1, ..., N-1 }.                                                //
     //                                                                          //
     // Note: this functions is used to prepare a vector to be multiplied by the //
     // continuous-time normalized descriptor (static variable 'CND::cnd').      //
     //--------------------------------------------------------------------------//
     // Implementation note: This function uses two types of coordinate for      //
     // global states. The coordinates using the variable 'i' concern the rank   //
     // of the global states in the original model. The coordinates using the    //
     // variable 'j' concern only the "existing" states. After a process of      //
     // grouping some global states can be eliminated, the static variable       //
     // 'RSS::rss' keeps the information of which global states are "existing",  //
     // i.e., were not eliminated. All coordinates to the functions of the class //
     // 'RSS::rss' must be in the coordinates of the original model. However the //
     // structures 'VCT::vct' and 'CND::cnd' only accepts the coordinates of the //
     // current model (the coordinates of the "remaining states - see 'rss.h').  //
     //--------------------------------------------------------------------------//
void VCT::init(const v_typ o)
{
#ifdef _PEPS_DEBUG_
  if (CND::cnd.ss() != RSS::rss.size())
    Programming_Error("incompatibility between RSS and CND", 2350)
#endif
 int i;             // counter for all states
 int j;             // counter for the remaining states
 rp  factor;        // a factor to normalize the vector elements
 rp  acc = zero;    // an accumulator to normalize the vector elements

 create(CND::cnd.ss());                // allocates the vector elements
 p_order.create(CND::cnd.aut());       // allocates the 'p_order' list object
 p_order.put();                        // fills 'permutation' with 0, 1, ... N-1
 CND::cnd.get_oj(p_jumps);             // copy the original jumps
 CND::cnd.get_s(sizes);                // copy the size of the automata
 switch (o) {
   case eq_vec :
     factor = one / rp(RSS::rss.reach_size());    // computes the probability
     for (i=0, j=0; i<RSS::rss.orig_size(); i++)  // for all states
       if (RSS::rss.remains(i)) {                    // if it was not eliminated
         if (RSS::rss.has(i))                           // if it is reachable
           vec[j] = factor;                                // puts the probability
         else                                           // if it is not reachable
           vec[j] = zero;                                  // puts zero
         j++;                                           // the next remaining state
       }
     break;
   case ap_vec :
     for (i=0, j=0; i<RSS::rss.orig_size(); i++)  // for all states
       if (RSS::rss.remains(i)) {                    // if it was not eliminated
         if (RSS::rss.has(i)) {                         // if it is reachable
           vec[j] = one / CND::cnd.get_diag(j);            // puts the inverse of the
           acc += vec[j];                                  // corresp. diagonal element
         }                                                 // and add to the accumulator
         else                                           // if it is not reachable
           vec[j] = zero;                                  // puts zero
         j++;                                           // the next remaining state
       }
       factor = one / acc;             // normalizes the vector
       for (j=0; j<dimension; j++)
         vec[j] *= factor;
 }
}

     //--------------------------------------------------------------------------//
     //       It permutes the vector elements acording to a lexical order        //
     //--------------------------------------------------------------------------//
     //   The vector elements are sorted according to a lexical order expressed  //
     // by the list of automata 'permutation'. This function changes from the    //
     // current order (expressed in the variables 'p_order' and 'p_jumps') to a  //
     // new order (expressed by the input parameters 's' and 'j'). The vector    //
     // elements are permuted acording to this new list.                         //
     //--------------------------------------------------------------------------//
void VCT::permut(const aut_list & s, const jumps & j)
{
#ifdef _PEPS_DEBUG_
  if (!(s.is_full()))
    Programming_Error("incomplete list", 2351)
  if (s.size() != p_order.size())
    Programming_Error("different list sizes", 2352)
  if (j.size() != p_jumps.size())
    Programming_Error("different jump sizes", 2353)
#endif
 rp     * new_vec;
 aut_st   st;
 int      idx = 0;

 new_vec = new rp[dimension];
 st.create(sizes.size());
 st.reset();
 for (int i=0; i<dimension; i++) {
   new_vec[idx] = vec[i];
   st.next(p_order, sizes, j, idx);
 }
 p_order.copy(s);
 p_jumps.copy(j);
 delete[] vec;
 vec = new_vec;
}
     //--------------------------------------------------------------------------//
     //          It copies one vector elements to another (permuting)            //
     //--------------------------------------------------------------------------//
void VCT::becomes(const VCT & v, const aut_list & s, const jumps & j)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2354)
  if (!(s.is_full()))
    Programming_Error("incomplete list", 2355)
  if (s.size() != p_order.size())
    Programming_Error("different list sizes", 2356)
  if (j.size() != p_jumps.size())
    Programming_Error("different jump sizes", 2357)
#endif
 aut_st   st;
 int      idx = 0;

 sizes.copy(v.sizes);
 st.create(sizes.size());
 st.reset();
 for (int i=0; i<dimension; i++) {
   vec[idx] = v.vec[i];
   st.next(v.p_order, sizes, j, idx);
 }
 p_order.copy(s);
 p_jumps.copy(j);
}

     //--------------------------------------------------------------------------//
     //           It adds one vector elements to another (permuting)             //
     //--------------------------------------------------------------------------//

void VCT::operator += (const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2358)
#endif
 aut_st   st;
 int      idx = 0;

 st.create(sizes.size());
 st.reset();
 for (int i=0; i<dimension; i++) {
   vec[idx] += v.vec[i];
   st.next(v.p_order, sizes, p_jumps, idx);
 }
}

     //--------------------------------------------------------------------------//
     //           It adds one vector elements to another (normalizing)           //
     //--------------------------------------------------------------------------//
     //             Second version - normalizes the biggest residue              //
     //--------------------------------------------------------------------------//
void VCT::norm2_sum(const VCT & v)         //***************************************
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2359)
#endif
 int  k;
 rp   factor, min_factor = one;

 for (k=0; k<dimension; k++)               // searches the minimal factor
   if ((v.vec[k] + vec[k]) < zero) {
     factor = v.vec[k] / -vec[k];             // if the result will be negative
     if (factor < min_factor)
       min_factor = factor;
   }
   else
     if ((v.vec[k] + vec[k]) > one) {         // if the result will be bigger than one
       factor = (one - v.vec[k]) / vec[k];
       if (factor < min_factor)
         min_factor = factor;
     }
 scale(min_factor);                        // normalizes the residue ('this')
 for (k=0; k<dimension; k++)               // it performs   'this = v + this'
   vec[k] = v.vec[k] + vec[k];
 if (factor != one)                        // if a normalization was performed
   cout << one / factor << " norm. factor (" << factor << ")\n";  // inform it!
}

     //--------------------------------------------------------------------------//
     //           It adds one vector elements to another (normalizing)           //
     //--------------------------------------------------------------------------//
     //               First version - cut negatives and normalizes               //
     //--------------------------------------------------------------------------//
void VCT::norm_sum(const VCT & v)
{
#ifdef _PEPS_DEBUG_
  if (dimension != v.dimension)
    Programming_Error("different sized vectors", 2360)
#endif
 int  k;
 rp   acc = zero;
 rp   factor;
 rp   max_cut = zero;
 rp   total_cut = zero;
 int  cuts = 0;

 for (k=0; k<dimension; k++) {         // it performs   'this = v + this'
   vec[k] = v.vec[k] + vec[k];
   if (vec[k] < zero) {                   // if one of 'this' elements becomes negative
     if (absv(vec[k]) > max_cut)
       max_cut = absv(vec[k]);
     cuts++;
     total_cut += absv(vec[k]);
     vec[k] = zero;                          // "cuts" the element off (it becomes zero)
//   vec[k] = v.vec[k];               or     // keeps the old value
//   vec[k] = vec[k];                 or     // keeps it negative (do nothing)
   }
   acc += vec[k];                         // computes the sum of the elements of 'this'
 }
 factor = one / acc;                   // it normalizes 'this'
 for (k=0; k<dimension; k++)
   vec[k] *= factor;
 if (max_cut > close_to_zero)          // if at least a significant "cut" was performed
   cout << cuts << " cuts (" << max_cut << " - " << total_cut << ")\n";  // inform it!
}

     //--------------------------------------------------------------------------//
     //                      Output on 'fout' file stream                        //
     //--------------------------------------------------------------------------//
void VCT::write() const
{
 ofstream fout;
 Open_File(file_name, vct_file, fout);
 fout << dimension << "\n";
 sizes.write(fout);
 p_order.write(fout);
 p_jumps.write(fout);
 for (int i=0; i<dimension; i++)
   fout << vec[i] << "\n";
 Close_File(file_name, vct_file, fout);
}
     //--------------------------------------------------------------------------//
     //                      Input from 'fin' file stream                        //
     //--------------------------------------------------------------------------//
void VCT::read(const peps_name file)
{
 ifstream fin;
 int d;

 Open_File(file, vct_file, fin);
 fin >> d;
 create(d);
 baptise(file);
 sizes.read(fin);
 p_order.read(fin);
 p_jumps.read(fin);
 for (d=0; d<dimension; d++)
   fin >> vec[d];
 Close_File(file_name, vct_file, fin);
}
     //--------------------------------------------------------------------------//
     //              Memory space used by 'this' (in memory units)               //
     //--------------------------------------------------------------------------//
int VCT::mem_use() const
{
 int answer = 4;
 answer += dimension * 8;
 answer += p_order.mem_use();
 answer += p_jumps.mem_use();
 answer += sizes.mem_use();
 return(answer);
}

     //--------------------------------------------------------------------------//
     //                       Static Variables Allocation                        //
     //--------------------------------------------------------------------------//
     //    It allocates the internal static variables 't', 'z_in', 'z_out' and   //
     //    'g_st' according to the current values of the 'CND::cnd' structure    //
     //--------------------------------------------------------------------------//
void VCT::alloc_static()
{
 t.create(CND::cnd.ss());                   // creation and initialization of
 t.get_CND();                               // the temporary vector t (static)
 z_in  = new rp[CND::cnd.largest_aut()];    // allocation of the input "slice"
 z_out = new rp[CND::cnd.largest_aut()];    // allocation of the output "slice"
 g_st.create(CND::cnd.aut());               // allocation of the global state (CND size)
}

     //--------------------------------------------------------------------------//
     //                      Static Variables Desallocation                      //
     //--------------------------------------------------------------------------//
     //   It desallocates the internal static variables 't, z_in, z_out, g_st'   //
     //--------------------------------------------------------------------------//
void VCT::desalloc_static()
{
 t.erase();                                 // desallocates all temporary structures
 delete[] z_in;
 delete[] z_out;
 g_st.erase();
}
